@@ -7,14 +7,16 @@ SQLObject
 SQLObject is a object-relational mapper.  See SQLObject.html or
 SQLObject.txt for more.
 
-Modified by
+Modified by:
 
-    Daniel Savard, Xsoli Inc <sqlobject xsoli.com> 7 Feb 2004
-    - Added support for simple table inheritance.
+* Daniel Savard, Xsoli Inc <sqlobject xsoli.com> 7 Feb 2004
+  Added support for simple table inheritance.
 
-    Oleg Broytmann, SIA "ANK" <phd@phd.pp.ru> 3 Feb 2005
-    - Split inheritance support into a number of separate modules and classes -
-      InheritableSQLObject at al.
+* Oleg Broytmann, SIA "ANK" <phd@phd.pp.ru> 3 Feb 2005
+  Split inheritance support into a number of separate modules and classes -
+  InheritableSQLObject at al.
+
+* And others...
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as
@@ -33,6 +35,7 @@ USA.
 """
 
 import threading
+import weakref
 import sqlbuilder
 import dbconnection
 import col
@@ -43,6 +46,7 @@ import joins
 import index
 import classregistry
 import declarative
+import events
 from sresults import SelectResults
 from formencode import schema, compound
 
@@ -128,6 +132,11 @@ def findDependencies(name, registry=None):
     for klass in classregistry.registry(registry).allClasses():
         if findDependantColumns(name, klass):
             depends.append(klass)
+        else:
+            for join in klass.sqlmeta.joins:
+                if isinstance(join, joins.SORelatedJoin) and join.otherClassName == name:
+                    depends.append(klass)
+                    break
     return depends
 
 def findDependantColumns(name, klass):
@@ -177,6 +186,7 @@ class sqlmeta(object):
 
     table = None
     idName = None
+    idSequence = None
     # This function is used to coerce IDs into the proper format,
     # so you should replace it with str, or another function, if you
     # aren't using integer IDs
@@ -210,7 +220,7 @@ class sqlmeta(object):
     __metaclass__ = declarative.DeclarativeMeta
 
     # These attributes shouldn't be shared with superclasses:
-    _unshared_attributes = ['table', 'idName', 'columns']
+    _unshared_attributes = ['table', 'columns', 'childName']
 
     # These are internal bookkeeping attributes; the class-level
     # definition is a default for the instances, instances will
@@ -230,13 +240,24 @@ class sqlmeta(object):
     # will be true.  It's false by default:
     _perConnection = False
 
+    # Inheritance definitions:
+    parentClass = None # A reference to the parent class
+    childClasses = {} # References to child classes, keyed by childName
+    childName = None # Class name for inheritance child object creation
+
     def __classinit__(cls, new_attrs):
         for attr in cls._unshared_attributes:
             if not new_attrs.has_key(attr):
                 setattr(cls, attr, None)
+        declarative.setup_attributes(cls, new_attrs)
 
     def __init__(self, instance):
-        self.instance = instance
+        self.instance = weakref.proxy(instance)
+
+    def send(cls, signal, *args, **kw):
+        events.send(signal, cls.soClass, *args, **kw)
+
+    send = classmethod(send)
 
     def setClass(cls, soClass):
         cls.soClass = soClass
@@ -287,6 +308,9 @@ class sqlmeta(object):
     ########################################
 
     def addColumn(cls, columnDef, changeSchema=False, connection=None):
+        post_funcs = []
+        cls.send(events.AddColumnSignal, cls.soClass, connection,
+                 columnDef.name, columnDef, changeSchema, post_funcs)
         sqlmeta = cls
         soClass = cls.soClass
         del cls
@@ -299,8 +323,18 @@ class sqlmeta(object):
             "The class %s.%s already has a column %r (%r), you cannot "
             "add the column %r"
             % (soClass.__module__, soClass.__name__, name,
-               sqlmeta.columnDefinitions[name],
-               columnDef))
+               sqlmeta.columnDefinitions[name], columnDef))
+        # Collect columns from the parent classes to test
+        # if the column is not in a parent class
+        parent_columns = []
+        for base in soClass.__bases__:
+            if hasattr(base, "sqlmeta"):
+                parent_columns.extend(base.sqlmeta.columns.keys())
+        if hasattr(soClass, name):
+            assert  (name in parent_columns) or (name == "childName"), (
+                "The class %s.%s already has a variable or method %r, you cannot "
+                "add the column %r"
+                % (soClass.__module__, soClass.__name__, name, name))
         sqlmeta.columnDefinitions[name] = columnDef
         sqlmeta.columns[name] = column
         # A stable-ordered version of the list...
@@ -362,10 +396,10 @@ class sqlmeta(object):
         # SQLObject instance.
         if column.foreignKey:
 
-            # We go through the standard _SO_get_columnName
-            # deal, except chopping off the "ID" ending since
+            # We go through the standard _SO_get_columnName deal
             # we're giving the object, not the ID of the
             # object this time:
+            origName = column.origName
             if sqlmeta.cacheValues:
                 # self._SO_class_className is a reference
                 # to the class in question.
@@ -373,30 +407,22 @@ class sqlmeta(object):
             else:
                 # Same non-caching version as above.
                 getter = eval('lambda self: self._SO_foreignKey(self._SO_getValue(%s), self._SO_class_%s)' % (repr(name), column.foreignKey))
-            if column.origName.upper().endswith('ID'):
-                origName = column.origName[:-2]
-            else:
-                origName = column.origName
             setattr(soClass, rawGetterName(origName), getter)
 
             # And we set the _get_columnName version
-            # (sans ID ending)
-            if not hasattr(soClass, getterName(name)[:-2]):
-                setattr(soClass, getterName(name)[:-2], getter)
-                sqlmeta._plainForeignGetters[name[:-2]] = 1
+            if not hasattr(soClass, getterName(origName)):
+                setattr(soClass, getterName(origName), getter)
+                sqlmeta._plainForeignGetters[origName] = 1
 
             if not column.immutable:
                 # The setter just gets the ID of the object,
                 # and then sets the real column.
                 setter = eval('lambda self, val: setattr(self, %s, self._SO_getID(val))' % (repr(name)))
-                setattr(soClass, rawSetterName(name)[:-2], setter)
-                if not hasattr(soClass, setterName(name)[:-2]):
-                    setattr(soClass, setterName(name)[:-2], setter)
-                    sqlmeta._plainForeignSetters[name[:-2]] = 1
+                setattr(soClass, rawSetterName(origName), setter)
+                if not hasattr(soClass, setterName(origName)):
+                    setattr(soClass, setterName(origName), setter)
+                    sqlmeta._plainForeignSetters[origName] = 1
 
-            # We'll need to put in a real reference at
-            # some point.  See needSet at the top of the
-            # file for more on this.
             classregistry.registry(sqlmeta.registry).addClassCallback(
                 column.foreignKey,
                 lambda foreign, me, attr: setattr(me, attr, foreign),
@@ -413,6 +439,9 @@ class sqlmeta(object):
         if soClass._SO_finishedClassCreation:
             makeProperties(soClass)
 
+        for func in post_funcs:
+            func(soClass, column)
+
     addColumn = classmethod(addColumn)
 
     def addColumnsFromDatabase(sqlmeta, connection=None):
@@ -420,11 +449,14 @@ class sqlmeta(object):
         conn = connection or soClass._connection
         for columnDef in conn.columnsFromSchema(sqlmeta.table, soClass):
             if columnDef.name not in sqlmeta.columnDefinitions:
+                if isinstance(columnDef.name, unicode):
+                    columnDef.name = columnDef.name.encode('ascii')
                 sqlmeta.addColumn(columnDef)
 
     addColumnsFromDatabase = classmethod(addColumnsFromDatabase)
 
-    def delColumn(sqlmeta, column, changeSchema=False, connection=None):
+    def delColumn(cls, column, changeSchema=False, connection=None):
+        sqlmeta = cls
         soClass = sqlmeta.soClass
         if isinstance(column, str):
             column = sqlmeta.columns[column]
@@ -436,6 +468,9 @@ class sqlmeta(object):
             else:
                 raise IndexError(
                     "Column with definition %r not found" % column)
+        post_funcs = []
+        cls.send(events.DeleteColumnSignal, connection, column.name, column,
+                 post_funcs)
         name = column.name
         del sqlmeta.columns[name]
         del sqlmeta.columnDefinitions[name]
@@ -447,22 +482,25 @@ class sqlmeta(object):
         if sqlmeta._plainSetters.has_key(name):
             delattr(soClass, setterName(name))
         if column.foreignKey:
-            delattr(soClass, rawGetterName(name)[:-2])
-            if sqlmeta._plainForeignGetters.has_key(name[:-2]):
-                delattr(soClass, getterName(name)[:-2])
-            delattr(soClass, rawSetterName(name)[:-2])
-            if sqlmeta._plainForeignSetters.has_key(name[:-2]):
-                delattr(soClass, setterName(name)[:-2])
+            delattr(soClass, rawGetterName(soClass.sqlmeta.style.instanceIDAttrToAttr(name)))
+            if sqlmeta._plainForeignGetters.has_key(name):
+                delattr(soClass, getterName(name))
+            delattr(soClass, rawSetterName(soClass.sqlmeta.style.instanceIDAttrToAttr(name)))
+            if sqlmeta._plainForeignSetters.has_key(name):
+                delattr(soClass, setterName(name))
         if column.alternateMethodName:
             delattr(soClass, column.alternateMethodName)
 
         if changeSchema:
             conn = connection or soClass._connection
-            conn.delColumn(sqlmeta.table, column)
+            conn.delColumn(sqlmeta, column)
 
         if soClass._SO_finishedClassCreation:
             unmakeProperties(soClass)
 
+        for func in post_funcs:
+            func(soClass, column)
+
     delColumn = classmethod(delColumn)
 
     ########################################
@@ -522,6 +560,10 @@ class sqlmeta(object):
     def delJoin(sqlmeta, joinDef):
         soClass = sqlmeta.soClass
         for join in sqlmeta.joins:
+            # previously deleted joins will be None, so it must
+            # be skipped or it'll error out on the next line.
+            if join is None:
+                continue
             if joinDef is join.joinDef:
                 break
         else:
@@ -560,8 +602,36 @@ class sqlmeta(object):
         cls.indexDefinitions.append(indexDef)
         index = indexDef.withClass(cls.soClass)
         cls.indexes.append(index)
+        setattr(cls.soClass, index.name, index)
     addIndex = classmethod(addIndex)
 
+    ########################################
+    ## Utility methods
+    ########################################
+
+    def asDict(self):
+        """
+        Return the object as a dictionary of columns to values.
+        """
+        result = {}
+        for key in self.columns:
+            result[key] = getattr(self.instance, key)
+        result['id'] = self.instance.id
+        return result
+
+    def expireAll(sqlmeta, connection=None):
+        """
+        Expire all instances of this class.
+        """
+        soClass = sqlmeta.soClass
+        connection = connection or soClass._connection
+        cache_set = connection.cache
+        cache_set.weakrefAll(soClass)
+        for item in cache_set.getAll(soClass):
+            item.expire()
+
+    expireAll = classmethod(expireAll)
+
 sqlhub = dbconnection.ConnectionHub()
 
 class _sqlmeta_attr(object):
@@ -585,9 +655,9 @@ class _sqlmeta_attr(object):
 warnings_level = 1
 exception_level = None
 # Current levels:
-#  1) Actively deprecated in version after 0.6.1 (0.7?); removed after
-#  2) Deprecated after 1 (0.8?)
-#  3) Deprecated after 2 (0.9?)
+#  1) Actively deprecated in version after 0.9; removed after
+#  2) Deprecated after 1 (0.9?)
+#  3) Deprecated after 2 (0.10?)
 
 def deprecated(message, level=1, stacklevel=2):
     if exception_level is not None and exception_level <= level:
@@ -595,6 +665,9 @@ def deprecated(message, level=1, stacklevel=2):
     if warnings_level is not None and warnings_level <= level:
         warnings.warn(message, DeprecationWarning, stacklevel=stacklevel)
 
+if sys.version_info[:3] < (2, 3, 0):
+    deprecated("Support for Python 2.2 has been declared obsolete and will be removed in the next release of SQLObject")
+
 def setDeprecationLevel(warning=1, exception=None):
     """
     Set the deprecation level for SQLObject.  Low levels are more
@@ -606,8 +679,8 @@ def setDeprecationLevel(warning=1, exception=None):
 
     The levels currently mean:
 
-      1) Deprecated in current version (0.7).  Will be removed in next
-         version (0.8)
+      1) Deprecated in current version (0.9).  Will be removed in next
+         version (0.10)
 
       2) Planned to deprecate in next version, remove later.
 
@@ -644,10 +717,11 @@ class SQLObject(object):
     #DSM: The _inheritable attribute controls wheter the class can by
     #DSM: inherited 'logically' with a foreignKey and a back reference.
     _inheritable = False # Is this class inheritable?
-    _parentClass = None # A reference to the parent class
     _parent = None # A reference to the parent instance
-    _childClasses = {} # References to child classes
     childName = None # Children name (to be able to get a subclass)
+    # moved to sqlmeta in 0.8:
+    _parentClass = _sqlmeta_attr('parentClass', 1)
+    _childClasses = _sqlmeta_attr('childClasses', 1)
 
     # The law of Demeter: the class should not call another classes by name
     SelectResultsClass = SelectResults
@@ -682,62 +756,6 @@ class SQLObject(object):
 
         cls._SO_finishedClassCreation = False
 
-        if '_columns' in new_attrs and not is_base:
-            deprecated(
-                'The _columns attribute you gave (%r) is deprecated; '
-                'columns should be added as class attributes'
-                % new_attrs['_columns'], level=1)
-            for column in new_attrs['_columns']:
-                if isinstance(column, str):
-                    column = col.Col(column)
-                implicitColumns.append(column)
-
-        if '_joins' in new_attrs and not is_base:
-            deprecated(
-                'The _joins attribute you gave (%r) is deprecated; '
-                'joins should be added as class attributes'
-                % new_attrs['_joins'], level=1)
-            for j in new_attrs['_joins']:
-                implicitJoins.append(j)
-
-
-        #DSM: Need to know very soon if the class is a child of an
-        #DSM: inheritable class. If so, we keep a link to our parent class.
-        cls._childClasses = {}
-        for superclass in cls.__bases__:
-            if hasattr(superclass, '_inheritable') and superclass._inheritable and \
-                superclass.__name__ <> "InheritableSQLObject":
-                    cls._parentClass = superclass
-                    cls._parent = None
-                    superclass._childClasses[cls.__name__] = cls
-
-        #DSM: If this class is a child of a parent class, we need to do some
-        #DSM: attribute check and add a foreign key to the parent.
-        if cls._parentClass:
-            #DSM: First, look for invalid column name:
-            #DSM: reserved ones or same as a parent
-            parentCols = cls.sqlmeta.columns.keys()
-            for column in implicitColumns:
-                cname = column.name
-                if cname == 'childName':
-                    raise AttributeError, \
-                        "The column name 'childName' is reserved"
-                if cname in parentCols:
-                    raise AttributeError, "The column '%s' is already " \
-                        "defined in an inheritable parent" % cname
-            #DSM: Remove columns if inherited from an inheritable class
-            #DSM: as we don't want them.  All we want is a foreign key
-            #DSM: that points to our parent
-            cls.sqlmeta.columns = {}
-            cls.sqlmeta.columnList = []
-            cls.sqlmeta.columnDefinitions = {}
-        #DSM: If this is inheritable, add some default columns
-        #DSM: to be able to link to children
-        if hasattr(cls, '_inheritable') and cls._inheritable and \
-            cls.__name__ <> "InheritableSQLObject":
-            cls.sqlmeta.columnDefinitions['childName'] = col.StringCol(
-                name='childName', default=None)
-
         ######################################################
         # Set some attributes to their defaults, if necessary.
         # First we get the connection:
@@ -766,6 +784,10 @@ class SQLObject(object):
         for column in implicitColumns:
             cls.sqlmeta.addColumn(column)
 
+        # Now the class is in an essentially OK-state, so we can
+        # set up any magic attributes:
+        declarative.setup_attributes(cls, new_attrs)
+
         if cls.sqlmeta.fromDatabase:
             cls.sqlmeta.addColumnsFromDatabase()
 
@@ -786,36 +808,23 @@ class SQLObject(object):
         if not is_base:
             cls.q = sqlbuilder.SQLObjectTable(cls)
 
-            #DSM: If we are a child, get the q magic from the parent
-            currentClass = cls
-            while currentClass._parentClass:
-                currentClass = currentClass._parentClass
-                for column in currentClass.sqlmeta.columnDefinitions.values():
-                    if type(column) == col.ForeignKey: continue
-                    setattr(cls.q, column.name,
-                        getattr(currentClass.q, column.name))
-
         classregistry.registry(cls.sqlmeta.registry).addClass(cls)
 
-    _style = _sqlmeta_attr('style', 2)
-    _table = _sqlmeta_attr('table', 2)
-    _idName = _sqlmeta_attr('idName', 2)
-    _lazyUpdate = _sqlmeta_attr('lazyUpdate', 2)
-    _defaultOrder = _sqlmeta_attr('defaultOrder', 2)
-    _cacheValues = _sqlmeta_attr('cacheValues', 2)
-    _registry = _sqlmeta_attr('registry', 2)
-    _idType = _sqlmeta_attr('idType', 2)
-    _fromDatabase = _sqlmeta_attr('fromDatabase', 2)
+    _style = _sqlmeta_attr('style', 1)
+    _table = _sqlmeta_attr('table', 1)
+    _idName = _sqlmeta_attr('idName', 1)
+    _lazyUpdate = _sqlmeta_attr('lazyUpdate', 1)
+    _defaultOrder = _sqlmeta_attr('defaultOrder', 1)
+    _cacheValues = _sqlmeta_attr('cacheValues', 1)
+    _registry = _sqlmeta_attr('registry', 1)
+    _idType = _sqlmeta_attr('idType', 1)
+    _fromDatabase = _sqlmeta_attr('fromDatabase', 1)
     _expired = _sqlmeta_attr('expired', 2)
-    _columns = _sqlmeta_attr('columnList', 1)
-    _columnDict = _sqlmeta_attr('columns', 1)
-    addColumn = _sqlmeta_attr('addColumn', 2)
-    delColumn = _sqlmeta_attr('delColumn', 2)
-    addJoin = _sqlmeta_attr('addJoin', 2)
-    delJoin = _sqlmeta_attr('delJoin', 2)
-    addIndex = _sqlmeta_attr('addIndex', 2)
-    delIndex = _sqlmeta_attr('delIndex', 2)
-    getSchema = _sqlmeta_attr('getSchema', 2)
+    addColumn = _sqlmeta_attr('addColumn', 1)
+    delColumn = _sqlmeta_attr('delColumn', 1)
+    addJoin = _sqlmeta_attr('addJoin', 1)
+    delJoin = _sqlmeta_attr('delJoin', 1)
+    addIndex = _sqlmeta_attr('addIndex', 1)
 
     # @classmethod
     def _SO_setupSqlmeta(cls, new_attrs, is_base):
@@ -870,12 +879,12 @@ class SQLObject(object):
         """
         for attr in ['_table', '_lazyUpdate', '_style', '_idName',
                      '_defaultOrder', '_cacheValues', '_registry',
-                     '_idType', '_fromDatabase']:
+                     '_idType', '_fromDatabase', '_idSequence']:
             if new_attrs.has_key(attr):
                 new_name = attr[1:]
                 deprecated("%r is deprecated; please set the %r "
                            "attribute in sqlmeta instead" %
-                           (attr, new_name), level=2,
+                           (attr, new_name), level=1,
                            stacklevel=5)
                 setattr(cls.sqlmeta, new_name, new_attrs[attr])
                 delattr(cls, attr)
@@ -883,7 +892,7 @@ class SQLObject(object):
             if new_attrs.has_key(attr):
                 deprecated("%r is deprecated and read-only; please do "
                            "not use it in your classes until it is fully "
-                           "deprecated" % attr, level=3, stacklevel=5)
+                           "deprecated" % attr, level=2, stacklevel=5)
 
     _SO_cleanDeprecatedAttrs = classmethod(_SO_cleanDeprecatedAttrs)
 
@@ -909,6 +918,13 @@ class SQLObject(object):
                 cache.put(id, cls, val)
             finally:
                 cache.finishPut(cls)
+        elif selectResults and not val.dirty:
+            val._SO_writeLock.acquire()
+            try:
+                val._SO_selectInit(selectResults)
+                val.sqlmeta.expired = False
+            finally:
+                val._SO_writeLock.release()
         return val
 
     get = classmethod(get)
@@ -923,10 +939,6 @@ class SQLObject(object):
         # created, unlike __init__ which would be called
         # anytime the object was returned from cache.
         self.id = id
-        # We shadow the sqlmeta class with an instance of sqlmeta
-        # that points to us (our sqlmeta buddy object; where the
-        # sqlmeta class is our class's buddy class)
-        self.sqlmeta = self.__class__.sqlmeta(self)
         self._SO_writeLock = threading.Lock()
 
         # If no connection was given, we'll inherit the class
@@ -1016,6 +1028,7 @@ class SQLObject(object):
                 delattr(self, instanceName(column.name))
             self.sqlmeta.expired = True
             self._connection.cache.expire(self.id, self.__class__)
+            self._SO_createValues = {}
         finally:
             self._SO_writeLock.release()
 
@@ -1027,6 +1040,16 @@ class SQLObject(object):
         # in the database, and we can't insert it until all
         # the parts are set.  So we just keep them in a
         # dictionary until later:
+        d = {name: value}
+        if not self.sqlmeta._creating:
+            self.sqlmeta.send(events.RowUpdateSignal, self, d)
+        if len(d) != 1 or name not in d:
+            # Already called RowUpdateSignal, don't call it again
+            # inside .set()
+            self.sqlmeta.row_update_sig_suppress = True
+            self.set(**d)
+            del self.sqlmeta.row_update_sig_suppress
+        value = d[name]
         if from_python:
             dbValue = from_python(value, self._SO_validatorState)
         else:
@@ -1047,6 +1070,8 @@ class SQLObject(object):
             setattr(self, instanceName(name), value)
 
     def set(self, **kw):
+        if not self.sqlmeta._creating and not getattr(self.sqlmeta, "row_update_sig_suppress", False):
+            self.sqlmeta.send(events.RowUpdateSignal, self, kw)
         # set() is used to update multiple values at once,
         # potentially with one SQL statement if possible.
 
@@ -1072,6 +1097,9 @@ class SQLObject(object):
                 if to_python:
                     value = to_python(dbValue, self._SO_validatorState)
                 setattr(self, instanceName(name), value)
+
+            self._SO_createValues.update(kw)
+
             for name, value in extra.items():
                 try:
                     getattr(self.__class__, name)
@@ -1083,7 +1111,6 @@ class SQLObject(object):
                 except AttributeError, e:
                     raise AttributeError, '%s (with attribute %r)' % (e, name)
 
-            self._SO_createValues.update(kw)
             self.dirty = True
             return
 
@@ -1161,17 +1188,25 @@ class SQLObject(object):
             return joinClass.get(id)
 
     def __init__(self, **kw):
+        # We shadow the sqlmeta class with an instance of sqlmeta
+        # that points to us (our sqlmeta buddy object; where the
+        # sqlmeta class is our class's buddy class)
+        self.sqlmeta = self.__class__.sqlmeta(self)
         # The get() classmethod/constructor uses a magic keyword
         # argument when it wants an empty object, fetched from the
         # database.  So we have nothing more to do in that case:
         if kw.has_key('_SO_fetch_no_create'):
             return
 
+        post_funcs = []
+        self.sqlmeta.send(events.RowCreateSignal, kw, post_funcs)
+
         # Pass the connection object along if we were given one.
         if kw.has_key('connection'):
             self._connection = kw['connection']
             self.sqlmeta._perConnection = True
             del kw['connection']
+
         self._SO_writeLock = threading.Lock()
 
         if kw.has_key('id'):
@@ -1181,6 +1216,8 @@ class SQLObject(object):
             id = None
 
         self._create(id, **kw)
+        for func in post_funcs:
+            func(self)
 
     def _create(self, id, **kw):
 
@@ -1198,10 +1235,12 @@ class SQLObject(object):
                 default = column.default
 
                 # If we don't get it, it's an error:
-                if default is NoDefault:
+                # If we specified an SQL DEFAULT, then we should use that
+                if default is NoDefault and column.defaultSQL is not None:
                     raise TypeError, "%s() did not get expected keyword argument %s" % (self.__class__.__name__, column.name)
                 # Otherwise we put it in as though they did pass
                 # that keyword:
+
                 kw[column.name] = default
 
         self.set(**kw)
@@ -1235,23 +1274,39 @@ class SQLObject(object):
         cache = self._connection.cache
         cache.created(id, self.__class__, self)
         self._init(id)
+        post_funcs = []
+        kw = dict([('class', self.__class__), ('id', id)])
+        self.sqlmeta.send(events.RowCreatedSignal, kw, post_funcs)
+        for func in post_funcs:
+            func(self)
 
     def _SO_getID(self, obj):
         return getID(obj)
 
     def _findAlternateID(cls, name, dbName, value, connection=None):
+        if isinstance(value, unicode):
+            column = cls.sqlmeta.columns[name]
+            if isinstance(column, col.SOUnicodeCol):
+                value = value.encode(column.dbEncoding)
         return (connection or cls._connection)._SO_selectOneAlt(
             cls,
             [cls.sqlmeta.idName] +
-            [col.dbName for col in cls.sqlmeta.columnList],
+            [column.dbName for column in cls.sqlmeta.columnList],
             dbName,
             value), None
     _findAlternateID = classmethod(_findAlternateID)
 
-    def _SO_fetchAlternateID(cls, name, dbName, value, connection=None):
+    def _SO_fetchAlternateID(cls, name, dbName, value, connection=None, idxName=None):
         result, obj = cls._findAlternateID(name, dbName, value, connection)
         if not result:
-            raise SQLObjectNotFound, "The %s by alternateID %s=%s does not exist" % (cls.__name__, name, repr(value))
+            if idxName is None:
+                raise SQLObjectNotFound, "The %s by alternateID %s = %s does not exist" % (cls.__name__, name, repr(value))
+            else:
+                names = []
+                for i in xrange(len(name)):
+                    names.append("%s = %s" % (name[i], repr(value[i])))
+                names = ', '.join(names)
+                raise SQLObjectNotFound, "The %s by unique index %s(%s) does not exist" % (cls.__name__, idxName, names)
         if obj:
            return obj
         if connection:
@@ -1269,7 +1324,7 @@ class SQLObject(object):
                orderBy=NoDefault, limit=None,
                lazyColumns=False, reversed=False,
                distinct=False, connection=None,
-               join=None):
+               join=None, forUpdate=False):
         return cls.SelectResultsClass(cls, clause,
                              clauseTables=clauseTables,
                              orderBy=orderBy,
@@ -1278,10 +1333,15 @@ class SQLObject(object):
                              reversed=reversed,
                              distinct=distinct,
                              connection=connection,
-                             join=join)
+                             join=join, forUpdate=forUpdate)
     select = classmethod(select)
 
     def selectBy(cls, connection=None, **kw):
+        for key, column in cls.sqlmeta.columns.items():
+            if (key in kw) and isinstance(column, col.SOUnicodeCol):
+                value = kw[key]
+                if isinstance(value, unicode):
+                    kw[key] = value.encode(column.dbEncoding)
         conn = connection or cls._connection
         return cls.SelectResultsClass(cls,
                              conn._SO_columnClause(cls, kw),
@@ -1289,45 +1349,78 @@ class SQLObject(object):
 
     selectBy = classmethod(selectBy)
 
+    def tableExists(cls, connection=None):
+        conn = connection or cls._connection
+        return conn.tableExists(cls.sqlmeta.table)
+
+    tableExists = classmethod(tableExists)
+
     def dropTable(cls, ifExists=False, dropJoinTables=True, cascade=False,
                   connection=None):
         conn = connection or cls._connection
-        if ifExists and not conn.tableExists(cls.sqlmeta.table):
+        if ifExists and not cls.tableExists(connection=conn):
             return
+        extra_sql = []
+        post_funcs = []
+        cls.sqlmeta.send(events.DropTableSignal, cls, connection,
+                         extra_sql, post_funcs)
         conn.dropTable(cls.sqlmeta.table, cascade)
         if dropJoinTables:
             cls.dropJoinTables(ifExists=ifExists, connection=conn)
+        for sql in extra_sql:
+            connection.query(sql)
+        for func in post_funcs:
+            func(cls, conn)
     dropTable = classmethod(dropTable)
 
     def createTable(cls, ifNotExists=False, createJoinTables=True,
-                    createIndexes=True,
+                    createIndexes=True, applyConstraints=True,
                     connection=None):
         conn = connection or cls._connection
-        if ifNotExists and conn.tableExists(cls.sqlmeta.table):
+        if ifNotExists and cls.tableExists(connection=conn):
             return
-        conn.createTable(cls)
+        extra_sql = []
+        post_funcs = []
+        cls.sqlmeta.send(events.CreateTableSignal, cls, connection,
+                         extra_sql, post_funcs)
+        constraints = conn.createTable(cls)
+        if applyConstraints:
+            for constraint in constraints:
+                conn.query(constraint)
+        else:
+            extra_sql.extend(constraints)
         if createJoinTables:
             cls.createJoinTables(ifNotExists=ifNotExists,
                                  connection=conn)
         if createIndexes:
             cls.createIndexes(ifNotExists=ifNotExists,
                               connection=conn)
+        for func in post_funcs:
+            func(cls, conn)
+        return extra_sql
     createTable = classmethod(createTable)
 
-    def createTableSQL(cls, createJoinTables=True, connection=None,
-                       createIndexes=True):
+    def createTableSQL(cls, createJoinTables=True, createIndexes=True,
+                       connection=None):
         conn = connection or cls._connection
-        sql = conn.createTableSQL(cls)
+        sql, constraints = conn.createTableSQL(cls)
         if createJoinTables:
-            sql += '\n' + cls.createJoinTablesSQL(connection=conn)
+            join_sql = cls.createJoinTablesSQL(connection=conn)
+            if join_sql:
+                sql += ';\n' + join_sql
         if createIndexes:
-            sql += '\n' + cls.createIndexesSQL(connection=conn)
-        return sql
+            index_sql = cls.createIndexesSQL(connection=conn)
+            if index_sql:
+                sql += ';\n' + index_sql
+        return sql, constraints
     createTableSQL = classmethod(createTableSQL)
 
     def createJoinTables(cls, ifNotExists=False, connection=None):
         conn = connection or cls._connection
         for join in cls._getJoinsToCreate():
+            if not getattr(join, 'createRelatedTable', True):
+                # This join has requested not to be created
+                continue
             if (ifNotExists and
                 conn.tableExists(join.intermediateTable)):
                 continue
@@ -1338,8 +1431,11 @@ class SQLObject(object):
         conn = connection or cls._connection
         sql = []
         for join in cls._getJoinsToCreate():
+            if not getattr(join, 'createRelatedTable', True):
+                # This join has requested not to be created
+                continue
             sql.append(conn._SO_createJoinTableSQL(join))
-        return '\n'.join(sql)
+        return ';\n'.join(sql)
     createJoinTablesSQL = classmethod(createJoinTablesSQL)
 
     def createIndexes(cls, ifNotExists=False, connection=None):
@@ -1357,7 +1453,7 @@ class SQLObject(object):
             if not index:
                 continue
             sql.append(conn.createIndexSQL(cls, index))
-        return '\n'.join(sql)
+        return ';\n'.join(sql)
     createIndexesSQL = classmethod(createIndexesSQL)
 
     def _getJoinsToCreate(cls):
@@ -1400,19 +1496,39 @@ class SQLObject(object):
     clearTable = classmethod(clearTable)
 
     def destroySelf(self):
+        self.sqlmeta.send(events.RowDestroySignal, self)
         # Kills this object.  Kills it dead!
-        depends = []
+
         klass = self.__class__
+
+        # Free related joins on the base class
+        for join in klass.sqlmeta.joins:
+            if isinstance(join, joins.SORelatedJoin):
+                q = "DELETE FROM %s WHERE %s=%d" % (join.intermediateTable, join.joinColumn, self.id)
+                self._connection.query(q)
+
+        depends = []
         depends = self._SO_depends()
         for k in depends:
+            # Free related joins
+            for join in k.sqlmeta.joins:
+                if isinstance(join, joins.SORelatedJoin) and join.otherClassName == klass.__name__:
+                     q = "DELETE FROM %s WHERE %s=%d" % (join.intermediateTable, join.otherColumn, self.id)
+                     self._connection.query(q)
+
             cols = findDependantColumns(klass.__name__, k)
+
+            # Don't confuse the rest of the process
+            if len(cols) == 0:
+                continue
+
             query = []
             delete = setnull = restrict = False
             for col in cols:
                 if col.cascade == False:
                     # Found a restriction
                     restrict = True
-                query.append("%s = %s" % (col.dbName, self.id))
+                query.append(getattr(k.q, col.name) == self.id)
                 if col.cascade == 'null':
                     setnull = col.name
                 elif col.cascade:
@@ -1421,7 +1537,7 @@ class SQLObject(object):
                 "Class %s depends on %s accoriding to "
                 "findDependantColumns, but this seems inaccurate"
                 % (k, klass))
-            query = ' OR '.join(query)
+            query = sqlbuilder.OR(*query)
             results = k.select(query, connection=self._connection)
             if restrict:
                 if results.count():
@@ -1448,7 +1564,23 @@ class SQLObject(object):
 
     delete = classmethod(delete)
 
+    def deleteMany(cls, where=NoDefault, connection=None):
+        conn = connection or cls._connection
+        conn.query(conn.sqlrepr(sqlbuilder.Delete(cls.sqlmeta.table, where)))
+
+    deleteMany = classmethod(deleteMany)
+
+    def deleteBy(cls, connection=None, **kw):
+        conn = connection or cls._connection
+        conn.query(conn.sqlrepr(sqlbuilder.Delete(cls.sqlmeta.table,
+            conn._SO_columnClause(cls, kw))))
+
+    deleteBy = classmethod(deleteBy)
+
     def __repr__(self):
+        if not hasattr(self, 'id'):
+            # Object initialization not finished.  No attributes can be read.
+            return '<%s (not initialized)>' % self.__class__.__name__
         return '<%s %r %s>' \
                % (self.__class__.__name__,
                   self.id,
@@ -1483,6 +1615,7 @@ class SQLObject(object):
         cls._connection = value
     setConnection = classmethod(setConnection)
 
+
 def capitalize(name):
     return name[0].capitalize() + name[1:]
 
@@ -1501,7 +1634,7 @@ def instanceName(name):
 class SQLObjectState(object):
 
     def __init__(self, soObject):
-        self.soObject = soObject
+        self.soObject = weakref.proxy(soObject)
         self.protocol = 'sql'
 
 
