@@ -55,7 +55,7 @@ def test_string_000():
     assert sqlrepr('A String\000Another', 'postgres') == "'A String\\0Another'"
 
 def test_string_():
-    assert sqlrepr('A String\'Another', 'postgres') == "'A String\\'Another'"
+    assert sqlrepr('A String\tAnother', 'postgres') == "'A String\\tAnother'"
     assert sqlrepr('A String\'Another', 'firebird') == "'A String''Another'"
 
 def test_simple_unicode():
@@ -82,6 +82,15 @@ def test_bool():
     assert sqlrepr(TRUE, 'mysql') == "1"
     assert sqlrepr(FALSE, 'mysql') == "0"
 
+def test_datetime():
+    from datetime import datetime, date, time
+    assert sqlrepr(datetime(2005, 7, 14, 13, 31, 2)) == "'2005-07-14 13:31:02'"
+    assert sqlrepr(date(2005, 7, 14)) == "'2005-07-14'"
+    assert sqlrepr(time(13, 31, 2)) == "'13:31:02'"
+    # now dates before 1900
+    assert sqlrepr(datetime(1428, 7, 14, 13, 31, 2)) == "'1428-07-14 13:31:02'"
+    assert sqlrepr(date(1428, 7, 14)) == "'1428-07-14'"
+
 def test_instance():
     instance = TestClass()
     assert sqlrepr(instance) == repr(instance)
@@ -103,9 +112,39 @@ def test_select():
     assert sqlrepr(instance, 'mysql') == "SELECT 'test'"
 
 def test_insert():
+    # Single column, no keyword arguments.
     instance = Insert('test', [('test',)])
     assert sqlrepr(instance, 'mysql') == "INSERT INTO test VALUES ('test')"
 
+    # Multiple columns, no keyword arguments.
+    instance2 = Insert('test', [('1st', '2nd', '3th', '4th')])
+    assert sqlrepr(instance2, 'postgres') == "INSERT INTO test VALUES ('1st', '2nd', '3th', '4th')"
+
+    # Multiple rows, multiple columns, "valueList" keyword argument.
+    instance3 = Insert('test', valueList=[('a1', 'b1'), ('a2', 'b2'), ('a3', 'b3')])
+    assert sqlrepr(instance3, 'sqlite') == "INSERT INTO test VALUES ('a1', 'b1'), ('a2', 'b2'), ('a3', 'b3')"
+
+    # Multiple columns, "values" keyword argument.
+    instance4 = Insert('test', values=('v1', 'v2', 'v3'))
+    assert sqlrepr(instance4, 'mysql') == "INSERT INTO test VALUES ('v1', 'v2', 'v3')"
+
+    # Single column, "valueList" keyword argument.
+    instance5 = Insert('test', valueList=[('v1',)])
+    assert sqlrepr(instance5, 'mysql') == "INSERT INTO test VALUES ('v1')"
+
+    # Multiple rows, Multiple columns, template.
+    instance6 = Insert('test', valueList=[('a1', 'b1'), ('a2', 'b2')], template=['col1', 'col2'])
+    assert sqlrepr(instance6, 'mysql') == "INSERT INTO test (col1, col2) VALUES ('a1', 'b1'), ('a2', 'b2')"
+
+    # Multiple columns, implicit template (dictionary value).
+    instance7 = Insert('test', valueList=[{'col1': 'a1', 'col2': 'b1'}])
+    assert sqlrepr(instance7, 'mysql') == "INSERT INTO test (col2, col1) VALUES ('b1', 'a1')"
+
+    # Multiple rows, Multiple columns, implicit template.
+    instance8 = Insert('test', valueList=[{'col1': 'a1', 'col2': 'b1'},
+                                        {'col1': 'a2', 'col2': 'b2'}])
+    assert sqlrepr(instance8, 'mysql') == "INSERT INTO test (col2, col1) VALUES ('b1', 'a1'), ('b2', 'a2')"
+
 def test_update():
     instance = Update('test', {'test':'test'})
     assert sqlrepr(instance, 'mysql') == "UPDATE test SET test='test'"
@@ -124,7 +163,7 @@ def test_trueclause():
 
 def test_op():
     instance = SQLOp('and', 'this', 'that')
-    assert sqlrepr(instance, 'mysql') == "('this' AND 'that')"
+    assert sqlrepr(instance, 'mysql') == "(('this') AND ('that'))"
 
 def test_call():
     instance = SQLCall('test', ('test',))
@@ -137,3 +176,20 @@ def test_constant():
 def test_prefix():
     instance = SQLPrefix('test', 'test')
     assert sqlrepr(instance, 'mysql') == "test 'test'"
+
+def test_dict():
+    assert sqlrepr({"key": "value"}, "sqlite") == "('key')"
+
+def test_sets():
+    try:
+        set
+    except NameError:
+        pass
+    else:
+        assert sqlrepr(set([1])) == "(1)"
+    try:
+        from sets import Set
+    except ImportError:
+        pass
+    else:
+        assert sqlrepr(Set([1])) == "(1)"
